import logging
import json
from flask import Request
from wslarkbot import AESCipher, Bot as Client
from .base import BaseBot, Message


class FeishuChatBot(BaseBot):

    def __init__(self, app_id='', app_secret='', encrypt_key='', verification_token=''):
        super().__init__(app_id)
        self.app_secret = app_secret
        self.encrypt_key = encrypt_key
        self.verification_token = verification_token
        self.client = Client(app_id=app_id, app_secret=app_secret, verification_token=verification_token, encrypt_key=encrypt_key)
        self.on('filter', lambda message: 'challenge' not in message)

    def run(self, message):
        logging.info('FeishuChatBot.run', message)
        # return 'reply ' + message['content']
        return self.trigger('text', message)

    def _decrypt_data(self, encrypt_key, encrypt_data):
        cipher = AESCipher(encrypt_key)
        return json.loads(cipher.decrypt_string(encrypt_data))

    def parse_message(self, content):
        # TODO self.client._validate(self.encrypt_key, {'body': conetne.json})
        if isinstance(content, Request):
            data = content.json
            if 'challenge' in data:
                return Message(**data)
            if 'encrypt' in data:
                data = self.client._decrypt_data(self.encrypt_key, data['encrypt'])
            return Message(**data)
        content = content if isinstance(content, str) else str(content)
        return super().parse_message(content)

    def send(self, message):
        logging.info('FeishuChatBot.send', message)
        print('send', message, type(message))
        print('reply_text', message.event.message.message_id)
        print('reply_text', message.result)
        self.client.reply_text(message.event.message.message_id, message.result)

    def send_text(self, fn=None):
        self.on('text', fn)





from .globals import _cv_broker, _cv_message, _cv_instance, current_broker


_sentinel = object()


class BrokerContext:
    def __init__(self, broker, event=None, sender=None):
        self.broker = broker
        self.event = event
        self.sender = sender
        self._cv_tokens = []

    def copy(self):
        return self.__class__(
            self.broker,
            event=self.event,
            sender=self.sender,
        )

    def push(self):
        self._cv_tokens.append(_cv_broker.set(self))

    def pop(self, exc=_sentinel):
        try:
            if len(self._cv_tokens) == 1:
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                # do_teardown_broker_context(exc)
        finally:
            ctx = _cv_broker.get()
            _cv_broker.reset(self._cv_tokens.pop())
        if ctx is not self:
            raise AssertionError(
                f"Popped wrong broker context. ({ctx!r} instead of {self!r})"
            )

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.pop(exc_value)


class MessageContext:
    def __init__(self, broker, app_id, message, bot=None):
        self.broker = broker
        self.app_id = app_id
        self.message = message
        if bot is None:
            bot = broker.get_bot(app_id)
        self.bot = bot
        self._cv_tokens = []

    def copy(self):
        return self.__class__(
            self.broker,
            self.app_id,
            self.message,
            bot=self.bot,
        )

    def push(self):
        broker_ctx = _cv_broker.get(None)
        if broker_ctx is None or broker_ctx.broker is not self.broker:
            broker_ctx = self.broker.broker_context()
            broker_ctx.push()
        else:
            broker_ctx = None
        self._cv_tokens.append((_cv_message.set(self), broker_ctx))

    def pop(self, exc=_sentinel):
        try:
            if len(self._cv_tokens) == 1:
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                # do_teardown_message_context(exc)
        finally:
            ctx = _cv_message.get()
            token, broker_ctx = self._cv_tokens.pop()
            _cv_message.reset(token)
            if broker_ctx is not None:
                broker_ctx.pop(exc)
            if ctx is not self:
                raise AssertionError(
                    f"Popped wrong message context. ({ctx!r} instead of {self!r})"
                )

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.pop(exc_value)


class InstanceContext:
    def __init__(self):
        self._cv_tokens = []

    def push(self):
        from .broker import BaseBroker
        if _cv_broker.get(None) is None and isinstance(self, BaseBroker):
            _cv_broker.set(self.broker_context())
        self._cv_tokens.append(_cv_instance.set(self))

    def pop(self, exc=_sentinel):
        try:
            if len(self._cv_tokens) == 1:
                if exc is _sentinel:
                    exc = sys.exc_info()[1]
                # do_teardown_instance_context(exc)
        finally:
            instance = _cv_instance.get(None)
            from .broker import BaseBroker
            # can start broker ouside with
            if not isinstance(instance, BaseBroker):
                _cv_instance.reset(self._cv_tokens.pop())
            if instance is not self:
                raise AssertionError(
                    f"Popped wrong instance context. ({instance!r} instead of {self!r})"
                )

    def __enter__(self):
        self.push()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.pop(exc_value)




from contextvars import ContextVar
from werkzeug.local import LocalProxy

"""
1. broker_ctx: broker+event+sender
2. message_ctx: broker+bot
2. instance_ctx: current_instance 用于写多个层级的时候，不显式使用with也能找到对应的层级
"""


_no_broker_msg = "Working outside of broker context."
_cv_broker = ContextVar("connectai.broker_ctx")
broker_ctx = LocalProxy(_cv_broker, unbound_message=_no_broker_msg)
current_broker = LocalProxy(_cv_broker, "broker", unbound_message=_no_broker_msg)
broker = LocalProxy(_cv_broker, "broker", unbound_message=_no_broker_msg)
event = LocalProxy(_cv_broker, "event", unbound_message=_no_broker_msg)
sender = LocalProxy(_cv_broker, "sender", unbound_message=_no_broker_msg)


_no_message_msg = "Working outside of message context."
_cv_message = ContextVar("connectai.message_ctx")
message_ctx = LocalProxy(_cv_message, unbound_message=_no_message_msg)
bot = LocalProxy(_cv_message, "bot", unbound_message=_no_message_msg)
current_bot = LocalProxy(_cv_message, "bot", unbound_message=_no_message_msg)

_no_instance_msg = "Working outside of instance context."
_cv_instance = ContextVar("connectai.instance_ctx")
current_instance = LocalProxy(_cv_instance, unbound_message=_no_instance_msg)




import logging
import json
from typing import Dict


class Message(Dict):

    def __getattr__(self, name):
        result = self[name]
        if 'message' == name and 'content' in result and isinstance(result['content'], str):
            try:
                result['content'] = json.loads(result['content'])
                # remove @
                for mention in result.get('mentions', []):
                    result['content']['text'] = result['content']['text'].replace(mention['key'] + ' ', '    ')
                # 移除@_all
                result['content']['text'] = result['content']['text'].replace('@_all ', '')
            except Exception as e:
                logging.debug(e)

        return Message(**result) if isinstance(result, dict) else result




import logging
import asyncio
from flask import Flask, request, jsonify
from functools import partial, wraps
from .globals import current_broker, _cv_broker, broker_ctx, _cv_instance
from .ctx import InstanceContext
from .message import Message


class BaseEventHandler(InstanceContext):

    def __init__(self, app_id='noop'):
        super().__init__()
        self.app_id = app_id
        current_broker.register_event(self)

    async def start(self, *args, **kwargs):
        raise NotImplementedError


class FeishuCallbackHandler(BaseEventHandler):

    def __init__(self, app_id='noop', prefix='/api/feishu', port=8888):
        super().__init__(app_id)
        self.prefix = prefix
        self.port = port

    def get_app(self):
        app = Flask(__name__)
        # save broker_ctx in current scope
        broker_ctx = _cv_broker.get(None)

        def event_handler(app_id):
            bot = broker_ctx.broker.get_bot(app_id)
            if bot:
                message = bot.parse_message(request)
                if message and bot.filter(message):
                    broker_ctx.broker.bot_queue.put_nowait((app_id, message))
                if 'challenge' in message:
                    return jsonify(message)
            return ''

        app.add_url_rule(f"{self.prefix}/<app_id>", 'event_handler', event_handler, methods=['POST'])
        return app

    async def start(self, port=None):
        app = self.get_app()
        app.run(port=port or self.port)


class NoopEventHandler(BaseEventHandler):

    async def start(self):
        while True:
            content = input('User: ')
            message = self.parse_message(content)
            current_broker.bot_queue.put_nowait((self.app_id, message))
            await asyncio.sleep(0.1)

    def parse_message(self, content):
        return Message(app_id=self.app_id, content=content)




import time
from flask import Request
from ..globals import current_broker
from ..ctx import InstanceContext
from ..message import Message


class BaseBot(InstanceContext):

    def __init__(self, app_id=''):
        super().__init__()
        self.app_id = app_id
        self.events = {}
        current_broker.register_bot(self)

    def parse_message(self, content):
        return Message(app_id=self.app_id, content=content)

    def filter(self, message):
        # 支持使用事件按顺序进行过滤，遇到失败的就返回False
        for fn in self.events.get('filter', []):
            if not fn(message):
                return False
        return True

    # def on(self, event, fn):
    #     self.events[event] = fn

    # def trigger(self, event, *args, **kwargs):
    #     if event in self.events:
    #         return self.events[event](*args, **kwargs)

    def on(self, event, fn):
        if event not in self.events:
            self.events[event] = []
        self.events[event].append(fn)

    def trigger(self, event, *args, **kwargs):
        events = self.events.get(event, [])
        if len(events) == 0:
            return
        # 单个事件支持多回调函数
        # 这里返回结果的时候，如果只有一个结果，就不返回数组，有多个结果就返回数组
        result = [fn(*args, **kwargs) for fn in events]
        return result[0] if len(result) == 1 else result






import logging
import queue
import asyncio
from functools import partial
from .globals import current_broker, current_instance, _cv_broker, bot
from .ctx import InstanceContext, BrokerContext, MessageContext


class BaseBroker(InstanceContext):

    def __init__(self):
        self.events = dict()
        super().__init__()

    def broker_context(self):
        return BrokerContext(self)

    def message_context(self, app_id, message):
        return MessageContext(self, app_id, message)

    def _launch_events(self):
        return [asyncio.get_event_loop().create_task(event.start()) for _, event in self.events.items()]

    def register_event(self, event):
        self.events[event.app_id] = event

    def _launch_bot_consumer(self):
        return self._launch_consumer('bot')

    def _launch_sender_consumer(self):
        return self._launch_consumer('message')

    def _launch_consumer(self, typ):
        def worker(queue):
            while True:
                app_id, m = queue.get()
                ctx = self.message_context(app_id, m)
                error = None
                try:
                    try:
                        ctx.push()
                        func = bot.run if 'bot' == typ else bot.send
                        if asyncio.iscoroutinefunction(func):
                            result = asyncio.get_event_loop().run_until_complete(func(m))
                        else:
                            result = func(m)
                        if result and 'bot' == typ:
                            m.update(result=result)
                            logging.info("debug result %r", result)
                            current_broker.message_queue.put_nowait((app_id, m))
                    except Exception as e:
                        error = e
                except:  # noqa: B001
                    error = sys.exc_info()[1]
                    raise
                finally:
                    queue.task_done()
                    ctx.pop(error)

        queue = self.bot_queue if 'bot' == typ else self.message_queue
        return asyncio.get_event_loop().run_in_executor(None, partial(worker, queue))

    def launch(self):
        if _cv_broker.get(None):
            self._launch()
        else:
            with self.broker_context():
                self._launch()


class QueueBroker(BaseBroker):
    def __init__(self, maxsize=0):
        super().__init__()
        self.bot_queue = queue.Queue(maxsize=maxsize)
        self.message_queue = queue.Queue(maxsize=maxsize)
        self.bots = dict()

    def _launch(self):
        tasks = [
            self._launch_bot_consumer(),
            self._launch_sender_consumer(),
            *self._launch_events()
        ]
        asyncio.get_event_loop().run_until_complete(asyncio.wait(tasks))

    def get_bot(self, app_id):
        return self.bots[app_id] if app_id in self.bots else None

    def register_bot(self, bot):
        self.bots[bot.app_id] = bot


MessageBroker = QueueBroker
